"""A module for base classes in NgenicPy."""

import json
import logging
from typing import Any

import httpx

from ..const import API_URL  # noqa: TID252
from ..exceptions import ApiException, ClientException  # noqa: TID252

_LOGGER = logging.getLogger(__package__)

try:
    from time import monotonic
except ImportError:
    from time import time as monotonic


"""
    Note that Tune system Nodes generally report data in intervals of five minutes,
    so there is no point in polling the API for new data at a higher rate.

    Referencing "Rate Limiting" in the Ngenic API documentation: https://developer.ngenic.se/#introduction
"""
CACHE_INTERVAL = 5.0 * 60


class NgenicBase:
    """Superclass for all models."""

    def __init__(
        self, session: httpx.AsyncClient = None, json_data: dict[str, Any] | None = None
    ) -> None:
        """Initialize our base object.

        :param session:
            (required) httpx client
        :param json_data:
            (required) Json representation of the concrete model
        """

        # store the httpx session on each object so we can reuse the connection
        self._session = session

        # backing json of the model
        self._json = json_data

        # poor-man's cache for get requests
        self._cache = {}

    def json(self) -> dict[str, Any]:
        """Get a json representation of the model.

        :return:

        """
        return self._json

    def uuid(self) -> str:
        """Get uuid attribute."""
        return self["uuid"]

    def __setitem__(self, attribute: str, data: Any):
        """Set an attribute in the model's JSON representation.

        :param attribute:
            (required) The attribute to set in the JSON representation
        :param data:
            (required) The data to set in the JSON representation
        """
        self._json[attribute] = data

    def __getitem__(self, attribute: str) -> Any:
        """Get an attribute from the model's JSON representation.

        :param attribute:
            (required) The attribute to get from the JSON representation
        :return:
            The value of the attribute
        :raises AttributeError:
            If the attribute is not found in the JSON representation
        """
        if attribute not in self._json:
            raise AttributeError(attribute)
        return self._json[attribute]

    def update(self) -> None:
        """Raise an exception as update is not allowed."""
        raise ClientException(f"Cannot update a '{self.__class__.__name__}'")

    def _parse(self, response: httpx.Response) -> Any:
        rsp_json = None

        if response is None:
            return None

        if response.status_code == 204:
            return None

        try:
            rsp_json = response.json()
        except ValueError:
            raise ApiException(
                f"Ngenic API return an invalid json body (status={response.status_code})",
            ) from ValueError

        return rsp_json

    def _new_instance(
        self, instance_class: type, json_data: Any, **kwargs
    ) -> type | list[type] | None:
        """Create a new model instance.

        :param class instance_class:
            (required) class of instance to initialize with jsosn
        :param dict json_data:
            (required) data to initialize the instance with
        :param kwargs:
            Additional data required by the instance type
        :return:
            new `instance_class` or `list(instance_class)`
        """
        if json_data is not None and (
            not isinstance(json_data, dict) and not isinstance(json_data, list)
        ):
            raise ClientException(
                "Invalid data to create new instance with (expected json)"
            )
        if not json_data:
            return None

        if isinstance(json_data, list):
            return [
                instance_class(session=self._session, json_data=inst_json, **kwargs)
                for inst_json in json_data
            ]
        return instance_class(self._session, json_data, **kwargs)

    def _parse_new_instance(
        self, url: str, instance_class: type, invalidate_cache: bool = False, **kwargs
    ) -> type | list[type] | None:
        """Get JSON from an URL and create a new instance of it.

        :param str url:
            (required) url to get instance data from
        :param type instance_class:
            (required) class of instance to initialize with parsed data
        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param kwargs:
            may contain additional args to the instance class
        :return:
            new `instance_class`
        :rtype:
            `instance_class`
        """

        ret_json = self._parse(self._get(url, invalidate_cache))
        return self._new_instance(instance_class, ret_json, **kwargs)

    async def _async_parse_new_instance(
        self, url: str, instance_class: type, invalidate_cache: bool = False, **kwargs
    ) -> type | list[type] | None:
        """Get JSON from an URL and create a new instance of it.

        :param str url:
            (required) url to get instance data from
        :param type instance_class:
            (required) class of instance to initialize with parsed data
        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param kwargs:
            may contain additional args to the instance class
        :return:
            new `instance_class`
        :rtype:
            `instance_class`
        """
        ret_json = self._parse(await self._async_get(url, invalidate_cache))
        return self._new_instance(instance_class, ret_json, **kwargs)

    def _request(self, method: str, url: str, *args, **kwargs) -> httpx.Response:
        """Make a HTTP request.
        This is the generic method for all requests, it will handle errors etc in a common way.

        :param str method:
            (required) HTTP method (i.e get, post, delete)
        :param args:
            Additional args to requests lib
        :param kwargs:
            Additional kwargs to requests lib
        :return:
            request
        """

        _LOGGER.debug("%s %s with %s %s", method.upper(), url, args, kwargs)

        res: httpx.Response | None = None
        try:
            if not isinstance(self._session, httpx.Client):
                raise ValueError("Cannot use sync methods when context is async")

            request_method = getattr(self._session, method)
            res = request_method(url, *args, **kwargs)

            # raise for e.g. 401
            res.raise_for_status()

            return res
        except httpx.HTTPError as exc:
            raise ClientException(
                self._get_error("A request exception occurred", res, parent_ex=exc)
            ) from exc
        except Exception as exc:
            raise ClientException(
                self._get_error("An exception occurred", res, parent_ex=exc)
            ) from exc

    async def _async_request(
        self, method: str, url: str, is_retry: bool, *args, **kwargs
    ) -> httpx.Response:
        """Make a HTTP request (async).

        This is the generic method for all requests, it will handle errors etc in a common way.

        :param str method:
            (required) HTTP method (i.e get, post, delete)
        :param bool is_retry:
            Indicator if this execution is a retry.
            This is a temporary fix for retrying broken connections.
        :param args:
            Additional args to requests lib
        :param kwargs:
            Additional kwargs to requests lib
        :return:
            response
        """

        _LOGGER.debug("%s %s with %s %s", method.upper(), url, args, kwargs)

        res: httpx.Response | None = None
        try:
            if not isinstance(self._session, httpx.AsyncClient):
                raise TypeError("Cannot use async methods when context is sync")  # noqa: TRY301

            request_method = getattr(self._session, method)
            res = await request_method(url, *args, **kwargs)

            # raise for e.g. 401
            res.raise_for_status()

            return res
        except httpx.CloseError as exc:
            if is_retry:
                # only retry once
                raise ClientException(
                    self._get_error("A request exception occurred", res, parent_ex=exc)
                ) from exc
            # retry request
            _LOGGER.debug(
                "Got a CloseError while trying to send request. Retry request once"
            )
            return await self._async_request(method, url, True, *args, **kwargs)
        except httpx.HTTPError as exc:
            raise ClientException(
                self._get_error("A request exception occurred", res, parent_ex=exc)
            ) from exc
        except Exception as exc:
            raise ClientException(
                self._get_error("An exception occurred", res, parent_ex=exc)
            ) from exc

    def _get_error(
        self, msg: str, res: httpx.Response, parent_ex: Exception | None = None
    ):
        server_msg: str = None
        if res is not None and res.status_code == 429:
            server_msg = f"Too many requests have been made, retry again after {res.headers['X-RateLimit-Reset']}"
        else:
            try:
                server_msg = res.json()["message"]
            except Exception:  # pylint: disable=broad-except
                if res is not None:
                    server_msg = f"{res.status_code} {res.reason_phrase}, {res.text}"
                elif parent_ex is not None:
                    if isinstance(parent_ex, httpx.ConnectTimeout):
                        server_msg = "Timed out connecting to ngenic server"
                    elif isinstance(parent_ex, httpx.ConnectTimeout):
                        server_msg = "Timed out sending request to ngenic server"
                    else:
                        server_msg = str(parent_ex)
                else:
                    server_msg = "Unknown error"

        if server_msg is None or server_msg == "":
            if res is not None:
                server_msg = f"{res.status_code} {res.reason_phrase}, {res.text}"
            else:
                server_msg = "Unknown error"

        return f"{msg}: {server_msg}"

    def _prehandle_write(self, data: Any, is_json: bool, **kwargs):
        headers = {}

        if is_json:
            data = json.dumps(data) if data is not None else data
            headers["Content-Type"] = "application/json"

        if "headers" in kwargs:
            # let caller override headers
            headers.update(kwargs.get("headers"))

        return (data, headers)

    def _delete(self, url: str):
        return self._request("delete", f"{API_URL}/{url}")

    def _async_delete(self, url: str):
        return self._async_request("delete", f"{API_URL}/{url}", False)

    def _get(self, url: str, invalidate_cache: bool = False):
        req_url = f"{API_URL}/{url}"
        now = monotonic()
        cache_key = hash(req_url)

        if invalidate_cache:
            _LOGGER.debug("Cache invalidated for %s", req_url)

        if not invalidate_cache and cache_key in self._cache:
            expiration, result = self._cache[cache_key]
            if expiration > now:
                _LOGGER.debug("Cache hit for %s", req_url)
                return result

        result = self._request("get", req_url)

        expiration = now + CACHE_INTERVAL
        self._cache[cache_key] = expiration, result

        return result

    async def _async_get(self, url: str, invalidate_cache: bool = False):
        url = f"{API_URL}/{url}"
        now = monotonic()
        cache_key = hash(url)

        if invalidate_cache:
            _LOGGER.debug("Cache invalidated for %s", url)

        if not invalidate_cache and cache_key in self._cache:
            expiration, result = self._cache[cache_key]
            if expiration > now:
                _LOGGER.debug("Cache hit for %s", url)
                return result

        result = await self._async_request("get", url, False)

        expiration = now + CACHE_INTERVAL
        self._cache[cache_key] = expiration, result

        return result

    def _post(self, url: str, data: Any = None, is_json: bool = True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, **kwargs)
        return self._request("post", f"{API_URL}/{url}", data=data, headers=headers)

    def _async_post(self, url: str, data: Any = None, is_json: bool = True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, **kwargs)
        return self._async_request(
            "post", f"{API_URL}/{url}", False, data=data, headers=headers
        )

    def _put(self, url: str, data: Any = None, is_json: bool = True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, **kwargs)
        return self._request("put", f"{API_URL}/{url}", data=data, headers=headers)

    def _async_put(self, url: str, data: Any = None, is_json: bool = True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, **kwargs)
        return self._async_request(
            "put", f"{API_URL}/{url}", False, data=data, headers=headers
        )



"""Tune model."""

from typing import Any

import httpx

from ..const import API_PATH  # noqa: TID252
from .base import NgenicBase
from .node import Node
from .room import Room
from .setpoint_schedule import SetpointSchedule


class Tune(NgenicBase):
    """Ngenic API Tune model."""

    def __init__(self, session: httpx.AsyncClient, json_data: dict[str, Any]) -> None:
        """Initialize the tune model."""
        super().__init__(session=session, json_data=json_data)

    def uuid(self) -> str:
        """Get the tune UUID."""

        # If a tune was fetched with the list API, it contains "tuneUuid"
        # If it was fetched directly (with UUID), it contains "uuid"
        try:
            return self["tuneUuid"]
        except AttributeError:
            return super().uuid()

    def rooms(self, invalidate_cache: bool = False) -> list[Room]:
        """List all Rooms associated with a Tune.
        A Room contains an indoor sensor.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of rooms
        :rtype:
            `list(~ngenic.models.room.Room)`
        """
        url = API_PATH["rooms"].format(tune_uuid=self.uuid(), room_uuid="")
        return self._parse_new_instance(
            url, Room, invalidate_cache, tune_uuid=self.uuid()
        )

    async def async_rooms(self, invalidate_cache: bool = False) -> list[Room]:
        """List all Rooms associated with a Tune (async).
        A Room contains an indoor sensor.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of rooms
        :rtype:
            `list(~ngenic.models.room.Room)`
        """
        url = API_PATH["rooms"].format(tune_uuid=self.uuid(), room_uuid="")
        return await self._async_parse_new_instance(
            url, Room, invalidate_cache, tune_uuid=self.uuid()
        )

    def room(self, room_uuid: str, invalidate_cache: bool = False) -> Room:
        """Get data about a Room.
        A Room contains an indoor sensor.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param str room_uuid:
            (required) room UUID
        :return:
            the room
        :rtype:
            `~ngenic.models.room.Room`
        """
        url = API_PATH["rooms"].format(tune_uuid=self.uuid(), room_uuid=room_uuid)
        return self._parse_new_instance(
            url, Room, invalidate_cache, tune_uuid=self.uuid()
        )

    async def async_room(self, room_uuid: str, invalidate_cache: bool = False) -> Room:
        """Get data about a Room (async).
        A Room contains an indoor sensor.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param str room_uuid:
            (required) room UUID
        :return:
            the room
        :rtype:
            `~ngenic.models.room.Room`
        """
        url = API_PATH["rooms"].format(tune_uuid=self.uuid(), room_uuid=room_uuid)
        return await self._async_parse_new_instance(
            url, Room, invalidate_cache, tune_uuid=self.uuid()
        )

    def nodes(self, invalidate_cache: bool = False) -> list[Node]:
        """List all Nodes associated with a Tune.
        A Node is a logical network entity.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of nodes
        :rtype:
            `list(~ngenic.models.node.Node)`
        """
        url = API_PATH["nodes"].format(tune_uuid=self.uuid(), node_uuid="")
        return self._parse_new_instance(
            url, Node, invalidate_cache, tune_uuid=self.uuid()
        )

    async def async_nodes(self, invalidate_cache: bool = False) -> list[Node]:
        """List all Nodes associated with a Tune (async).
        A Node is a logical network entity.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of nodes
        :rtype:
            `list(~ngenic.models.node.Node)`
        """
        url = API_PATH["nodes"].format(tune_uuid=self.uuid(), node_uuid="")
        return await self._async_parse_new_instance(
            url, Node, invalidate_cache, tune_uuid=self.uuid()
        )

    def node(self, node_uuid: str, invalidate_cache: bool = False) -> Node:
        """Get data about a Node (async).
        A Node is a logical network entity.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param str node_uuid:
            (required) node UUID
        :return:
            the node
        :rtype:
            `~ngenic.models.node.Node`
        """
        url = API_PATH["nodes"].format(tune_uuid=self.uuid(), node_uuid=node_uuid)
        return self._parse_new_instance(
            url, Node, invalidate_cache, tune_uuid=self.uuid()
        )

    async def async_node(self, node_uuid: str, invalidate_cache: bool = False) -> Node:
        """Get data about a Node (async).
        A Node is a logical network entity.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param str node_uuid:
            (required) node UUID
        :return:
            the node
        :rtype:
            `~ngenic.models.node.Node`
        """
        url = API_PATH["nodes"].format(tune_uuid=self.uuid(), node_uuid=node_uuid)
        return await self._async_parse_new_instance(
            url, Node, invalidate_cache, tune_uuid=self.uuid()
        )

    def setpoint_schedule(
        self, name: str, invalidate_cache: bool = False
    ) -> SetpointSchedule:
        """Fetch the setpoint schedule.

        :param str name:
            name of the setpoint schedule
        :param bool invalidate_cache:
            if the cache should be invalidated or not (default: False)
        :return:
            the setpoint schedule
        :rtype:
            `~ngenic.models.setpoint_schedule.SetPointSchedule`
        """

        url = API_PATH["setpoint_schedules"].format(tune_uuid=self.uuid())
        rsp_json = self._parse(self._get(url, invalidate_cache))

        for obj in rsp_json:
            if obj["name"] == name:
                return self._new_instance(SetpointSchedule, obj, tune_uuid=self.uuid())

        new_json = {
            "name": name,
            "autoTune": True,
            "lowestSetpoint": 12,
        }
        return self._new_instance(SetpointSchedule, new_json, tune_uuid=self.uuid())

    async def async_setpoint_schedule(
        self, name: str, invalidate_cache: bool = False
    ) -> SetpointSchedule:
        """Fetch the setpoint schedule (async).

        :param str name:
            name of the setpoint schedule
        :param bool invalidate_cache:
            if the cache should be invalidated or not (default: False)
        :return:
            the setpoint schedule
        :rtype:
            `~ngenic.models.setpoint_schedule.SetPointSchedule`
        """

        url = API_PATH["setpoint_schedules"].format(tune_uuid=self.uuid())
        rsp_json = self._parse(await self._async_get(url, invalidate_cache))

        for obj in rsp_json:
            if obj["name"] == name:
                return self._new_instance(SetpointSchedule, obj, tune_uuid=self.uuid())

        new_json = {
            "name": name,
            "autoTune": True,
            "lowestSetpoint": 12,
        }
        return self._new_instance(SetpointSchedule, new_json, tune_uuid=self.uuid())



"""Setpoint Schedule model for Ngenic Tune API."""

import logging
from datetime import UTC, datetime, timedelta
from typing import Any

import httpx

from ..const import API_PATH  # noqa: TID252
from .base import NgenicBase

_LOGGER = logging.getLogger(__name__)


class SetpointSchedule(NgenicBase):
    """Setpoint Schedule model for Ngenic Tune API."""

    def __init__(
        self,
        session: httpx.AsyncClient,
        json_data: dict[str, Any],
        tune_uuid: str,
        away_days: int = 60,
    ) -> None:
        """Initialize the model."""
        self._parent_tune_uuid = tune_uuid
        self._away_days = away_days

        super().__init__(session=session, json_data=json_data)

        now = datetime.now(UTC).replace(second=0, microsecond=0)

        self._active = (
            (now >= self.start_time() and now < self.end_time())
            if self.start_time() is not None and self.end_time() is not None
            else False
        )

    def start_time(self) -> datetime | None:
        """Get startTime attribute."""

        try:
            date_string = self["startTime"]
        except Exception:  # pylint: disable=broad-except
            date_string = None

        try:
            return datetime.fromisoformat(date_string).astimezone(UTC)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.debug("Failed to parse startTime: %s", date_string)
            return None

    def end_time(self) -> datetime | None:
        """Get endTime attribute."""

        try:
            date_string = self["endTime"]
        except Exception:  # pylint: disable=broad-except
            date_string = None

        try:
            return datetime.fromisoformat(date_string).astimezone(UTC)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.debug("Failed to parse endTime: %s", date_string)
            return None

    def name(self) -> str:
        """Get name attribute."""
        return self["name"]

    def active(self) -> bool:
        """Get active attribute."""
        return self._active

    def activate_away(self) -> None:
        """Set the away schedule as activated from now on."""
        self._active = True
        now = datetime.now(UTC).replace(second=0, microsecond=0)

        # set schedule startTime to datetime.now(UTC) in ISO 8601:2004
        self["startTime"] = now.isoformat()

        # set schedule endTime to datetime.now(UTC) + XX days in ISO 8601:2004
        self["endTime"] = (now + timedelta(days=self._away_days)).isoformat()

    def deactivate_away(self) -> None:
        """Set the away schedule as deactivated."""
        self._active = False
        self["startTime"] = None
        self["endTime"] = None

    def set_schedule(self, start_time: datetime, end_time: datetime) -> None:
        """Set custom schedule."""
        self._active = True
        self["startTime"] = start_time.astimezone(UTC).isoformat()
        self["endTime"] = end_time.astimezone(UTC).isoformat()

    def update(self) -> None:
        """Update this schedule with its current values."""

        url = API_PATH["setpoint_schedules"].format(tune_uuid=self._parent_tune_uuid)

        try:
            schedule_uuid = self.uuid()
        except Exception:  # pylint: disable=broad-except
            schedule_uuid = None

        if not self._active and schedule_uuid is None:
            _LOGGER.debug("Deactivated schedule with no UUID, skipping update")
        elif not self._active:
            self._delete(f"{url}/{schedule_uuid}")
        elif schedule_uuid is not None:
            self._put(f"{url}/{schedule_uuid}", data=self.json())
        else:
            self._post(url, data=self.json())

    async def async_update(self) -> None:
        """Update this schedule with its current values (async)."""

        url = API_PATH["setpoint_schedules"].format(tune_uuid=self._parent_tune_uuid)

        try:
            schedule_uuid = self.uuid()
        except Exception:  # pylint: disable=broad-except
            schedule_uuid = None

        if not self._active and schedule_uuid is None:
            _LOGGER.debug("Deactivated schedule with no UUID, skipping update")
        elif not self._active:
            await self._async_delete(f"{url}/{schedule_uuid}")
        elif schedule_uuid is not None:
            await self._async_put(f"{url}/{schedule_uuid}", data=self.json())
        else:
            await self._async_post(url, data=self.json())



"""This module contains the main interfaces to the API."""

import ssl

import httpx

from .const import API_PATH
from .models import NgenicBase, Tune

# 30sec for connect, 10sec elsewhere.
timeout = httpx.Timeout(10.0, connect=20.0)


def _create_local_ssl_context() -> ssl.SSLContext:
    """Create SSL context."""
    return ssl.create_default_context()


# The default SSLContext objects are created at import time
SSL_CONTEXT_LOCAL_API = _create_local_ssl_context()


class BaseClient(NgenicBase):
    """Base class for the Ngenic API client."""

    def tunes(self, invalidate_cache: bool = False):
        """Fetch all tunes

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of tunes
        :rtype:
            `list(~ngenic.models.tune.Tune)`
        """
        url = API_PATH["tunes"].format(tune_uuid="")
        return self._parse_new_instance(url, Tune, invalidate_cache)

    async def async_tunes(self, invalidate_cache: bool = False) -> list[Tune]:
        """Fetch all tunes (async)

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of tunes
        :rtype:
            `list(~ngenic.models.tune.Tune)`
        """
        url = API_PATH["tunes"].format(tune_uuid="")
        return await self._async_parse_new_instance(url, Tune, invalidate_cache)

    def tune(self, tune_uuid: str, invalidate_cache: bool = False) -> Tune:
        """Fetch a single tune

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param str tune_uuid:
            (required) tune UUID
        :return:
            the tune
        :rtype:
            `~ngenic.models.tune.Tune`
        """
        url = API_PATH["tunes"].format(tune_uuid=tune_uuid)
        return self._parse_new_instance(url, Tune, invalidate_cache)

    def async_tune(self, tune_uuid: str, invalidate_cache: bool = False) -> Tune:
        """Fetch a single tune

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :param str tune_uuid:
            (required) tune UUID
        :return:
            the tune
        :rtype:
            `~ngenic.models.tune.Tune`
        """
        url = API_PATH["tunes"].format(tune_uuid=tune_uuid)
        return self._async_parse_new_instance(url, Tune, invalidate_cache)


class Ngenic(BaseClient):
    """Ngenic API client."""

    def __init__(self, token):
        """Initialize the client.

        :param token:
            (required) OAuth2 bearer token
        """

        # this will be added to the HTTP Authorization header for each request
        self._token = token

        # this header will be added to each HTTP request
        self._auth_headers = {"Authorization": f"Bearer {self._token}"}

        session = httpx.Client(
            headers=self._auth_headers, timeout=timeout, verify=SSL_CONTEXT_LOCAL_API
        )

        # initializing this doesn't require a session or json
        super(Ngenic, self).__init__(session=session)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """Close the session if it was not created as a context manager"""
        self._session.close()


class AsyncNgenic(BaseClient):
    """Ngenic API client (async)."""

    def __init__(self, token):
        """Initialize the async client.

        :param token:
            (required) OAuth2 bearer token
        """

        # this will be added to the HTTP Authorization header for each request
        self._token = token

        # this header will be added to each HTTP request
        self._auth_headers = {"Authorization": f"Bearer {self._token}"}

        session = httpx.AsyncClient(
            headers=self._auth_headers, timeout=timeout, verify=SSL_CONTEXT_LOCAL_API
        )

        # initializing this doesn't require a session or json
        super(AsyncNgenic, self).__init__(session=session)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.async_close()

    async def async_close(self):
        """Close the session if it was not created as a context manager"""
        await self._session.aclose()



"""Exceptions for ngenicpy."""


class NgenicException(Exception):
    """Base exception class."""


class ApiException(NgenicException):
    """Exception from ngenic API."""


class ClientException(NgenicException):
    """Exception from library."""

    def __init__(self, msg: str) -> None:
        """Initialize the exception."""
        super().__init__(self, msg)
        self.msg = msg



"""Ngenic API measurement model."""

from enum import Enum
from typing import Any

import httpx

from .base import NgenicBase


class MeasurementType(Enum):
    """Measurement type enumeration.

    Undocumented in API.
    """

    UNKNOWN = "unknown"
    CONTROL_VALUE = "control_value_C"
    ENERGY = "energy_kWH"
    FLOW = "flow_litre_per_hour"
    HUMIDITY = "humidity_relative_percent"
    INLET_FLOW_TEMPERATURE = "inlet_flow_temperature_C"
    L1_CURRENT = "L1_current_A"
    L1_VOLTAGE = "L1_voltage_V"
    L2_CURRENT = "L2_current_A"
    L2_VOLTAGE = "L2_voltage_V"
    L3_CURRENT = "L3_current_A"
    L3_VOLTAGE = "L3_voltage_V"
    POWER = "power_kW"
    PROCESS_VALUE = "process_value_C"
    PRODUCED_ENERGY = "produced_energy_kWH"
    PRODUCED_POWER = "produced_power_kW"
    RETURN_TEMPERATURE = "return_temperature_C"
    SETPOINT_VALUE = "setpoint_value_C"
    TARGET_TEMPERATURE = "target_temperature_C"
    TEMPERATURE = "temperature_C"

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


class Measurement(NgenicBase):
    """Ngenic API measurement model."""

    def __init__(
        self,
        session: httpx.AsyncClient,
        json_data: dict[str, Any],
        measurement_type: MeasurementType,
    ) -> None:
        """Initialize the measurement model."""
        self._measurement_type = measurement_type

        super().__init__(session=session, json_data=json_data)

    def get_type(self) -> MeasurementType:
        """Get the measurement type.

        :return:
            measurement type
        :rtype:
            `~ngenic.models.measurement.MeasurementType`
        """
        return self._measurement_type



"""Constants for the NgenicPy library."""

from typing import Final

API_URL: Final[str] = "https://app.ngenic.se/api/v3"

API_PATH: Final[dict[str, str]] = {
    "tunes": "tunes/{tune_uuid}",
    "rooms": "tunes/{tune_uuid}/rooms/{room_uuid}",
    "nodes": "tunes/{tune_uuid}/gateway/nodes/{node_uuid}",
    "node_status": "tunes/{tune_uuid}/nodestatus",
    "measurements": "tunes/{tune_uuid}/measurements/{node_uuid}",
    "measurements_types": "tunes/{tune_uuid}/measurements/{node_uuid}/types",
    "measurements_latest": "tunes/{tune_uuid}/measurements/{node_uuid}/latest",
    "setpoint_schedules": "tunes/{tune_uuid}/setpointschedules",
}



"""Node model."""

import asyncio
from enum import Enum
from typing import Any

import httpx

from ..const import API_PATH  # noqa: TID252
from .base import NgenicBase
from .measurement import Measurement, MeasurementType
from .node_status import NodeStatus


class NodeType(Enum):
    """Node type enumeration."""

    UNKNOWN = -1
    SENSOR = 0
    CONTROLLER = 1
    GATEWAY = 2
    INTERNAL = 3
    ROUTER = 4

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


class Node(NgenicBase):
    """Ngenic API node model."""

    def __init__(
        self, session: httpx.AsyncClient, json_data: dict[str, Any], tune_uuid: str
    ) -> None:
        """Initialize the node model."""
        self._parent_tune_uuid = tune_uuid

        # A cache for measurement types
        self._measurement_types = None

        super().__init__(session=session, json_data=json_data)

    def get_type(self):
        """Get the node type."""
        return NodeType(self["type"])

    def measurement_types(
        self, invalidate_cache: bool = False
    ) -> list[MeasurementType]:
        """Get types of available measurements for this node (async).

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of measurement type enums
        :rtype:
            `list(~ngenic.models.measurement.MeasurementType)
        """
        if not self._measurement_types:
            url = API_PATH["measurements_types"].format(
                tune_uuid=self._parent_tune_uuid, node_uuid=self.uuid()
            )
            measurements = self._parse(self._get(url, invalidate_cache))
            self._measurement_types = [MeasurementType(m) for m in measurements]

        return self._measurement_types

    async def async_measurement_types(
        self, invalidate_cache: bool = False
    ) -> list[MeasurementType]:
        """Get types of available measurements for this node (async).

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of measurement type enums
        :rtype:
            `list(~ngenic.models.measurement.MeasurementType)
        """
        if not self._measurement_types:
            url = API_PATH["measurements_types"].format(
                tune_uuid=self._parent_tune_uuid, node_uuid=self.uuid()
            )
            measurements = self._parse(await self._async_get(url, invalidate_cache))
            self._measurement_types = [MeasurementType(m) for m in measurements]

        return self._measurement_types

    def measurement(
        self,
        measurement_type: MeasurementType,
        from_dt: str | None = None,
        to_dt: str | None = None,
        period: str | None = None,
        invalidate_cache: bool = False,
    ) -> Measurement | list[Measurement] | None:
        """Get measurement for a specific period.

        :param MeasurementType measurement_type:
            (required) type of measurement
        :param from_dt:
            (optional) from datetime (ISO 8601:2004)
        :param to_dt:
            (optional) to datetime (ISO 8601:2004)
        :param period:
            (optional) Divides measurement interval into periods,
            default is a single period over entire interval.
            (ISO 8601:2004 duration format)
        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            the measurement.
            if no data is available for the period, None will be returned.
        :rtype:
            `list(~ngenic.models.measurement.Measurement)`
        """
        if from_dt is None:
            url = API_PATH["measurements_latest"].format(
                tune_uuid=self._parent_tune_uuid, node_uuid=self.uuid()
            )
            url += f"?type={measurement_type.value}"
            return self._parse_new_instance(
                url, Measurement, invalidate_cache, measurement_type=measurement_type
            )
        url = API_PATH["measurements"].format(
            tune_uuid=self._parent_tune_uuid, node_uuid=self.uuid()
        )
        url += f"?type={measurement_type.value}&from={from_dt}&to={to_dt}"
        if period:
            url += f"&period={period}"
        return self._parse_new_instance(
            url, Measurement, invalidate_cache, measurement_type=measurement_type
        )

    async def async_measurement(
        self,
        measurement_type: MeasurementType,
        from_dt: str | None = None,
        to_dt: str | None = None,
        period: str | None = None,
        invalidate_cache: bool = False,
    ) -> Measurement | list[Measurement] | None:
        """Get measurement for a specific period (async).

        :param MeasurementType measurement_type:
            (required) type of measurement
        :param from_dt:
            (optional) from datetime (ISO 8601:2004)
        :param to_dt:
            (optional) to datetime (ISO 8601:2004)
        :param period:
            (optional) Divides measurement interval into periods,
            default is a single period over entire interval.
            (ISO 8601:2004 duration format)
        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            the measurement.
            if no data is available for the period, None will be returned.
        :rtype:
            `list(~ngenic.models.measurement.Measurement)`
        """
        if from_dt is None:
            url = API_PATH["measurements_latest"].format(
                tune_uuid=self._parent_tune_uuid, node_uuid=self.uuid()
            )
            url += f"?type={measurement_type.value}"
            return await self._async_parse_new_instance(
                url, Measurement, invalidate_cache, measurement_type=measurement_type
            )
        url = API_PATH["measurements"].format(
            tune_uuid=self._parent_tune_uuid, node_uuid=self.uuid()
        )
        url += f"?type={measurement_type.value}&from={from_dt}&to={to_dt}"
        if period:
            url += f"&period={period}"
        return await self._async_parse_new_instance(
            url, Measurement, invalidate_cache, measurement_type=measurement_type
        )

    def measurements(self, invalidate_cache: bool = False) -> list[Measurement]:
        """Get latest measurements for a Node.

        Usually, you can get measurements from a `NodeType.SENSOR` or `NodeType.CONTROLLER`.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of measurements (if supported by the node)
        :rtype:
            `list(~ngenic.models.measurement.Measurement)`
        """
        # get available measurement types for this node
        measurement_types = self.measurement_types(invalidate_cache)

        # remove types that doesn't support reading from latest API
        if MeasurementType.ENERGY in measurement_types:
            measurement_types.remove(MeasurementType.ENERGY)
        if MeasurementType.PRODUCED_ENERGY in measurement_types:
            measurement_types.remove(MeasurementType.PRODUCED_ENERGY)

        if len(measurement_types) == 0:
            return []

        # retrieve latest measurement for each type
        latest_measurements = list(
            self.measurement(t, None, None, None, invalidate_cache)
            for t in measurement_types
        )

        # remove None measurements (caused by measurement types returning empty response)
        return list(m for m in latest_measurements if m)

    async def async_measurements(
        self, invalidate_cache: bool = False
    ) -> list[Measurement]:
        """Get latest measurements for a Node (async).

        Usually, you can get measurements from a `NodeType.SENSOR` or `NodeType.CONTROLLER`.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a list of measurements (if supported by the node)
        :rtype:
            `list(~ngenic.models.measurement.Measurement)`
        """
        # get available measurement types for this node
        measurement_types = await self.async_measurement_types(invalidate_cache)

        # remove types that doesn't support reading from latest API
        if MeasurementType.ENERGY in measurement_types:
            measurement_types.remove(MeasurementType.ENERGY)
        if MeasurementType.PRODUCED_ENERGY in measurement_types:
            measurement_types.remove(MeasurementType.PRODUCED_ENERGY)

        if len(measurement_types) == 0:
            return []

        # retrieve latest measurement for each type
        return list(
            await asyncio.gather(
                *[
                    self.async_measurement(t, None, None, None, invalidate_cache)
                    for t in measurement_types
                ]
            )
        )

    def status(self, invalidate_cache: bool = False) -> NodeStatus | None:
        """Get status about this Node.

        There are no API for getting the status for a single node, so we
        will use the list API and find our node in there.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not
        :return:
            a status object or None if Node doesn't support status
        :rtype:
            `~ngenic.models.node_status.NodeStatus`
        """
        url = API_PATH["node_status"].format(tune_uuid=self._parent_tune_uuid)
        rsp_json = self._parse(self._get(url, invalidate_cache))

        for status_obj in rsp_json:
            if status_obj["nodeUuid"] == self.uuid():
                return self._new_instance(NodeStatus, status_obj, node_uuid=self.uuid())
        return None

    async def async_status(self, invalidate_cache: bool = False) -> NodeStatus | None:
        """Get status about this Node (async).

        There are no API for getting the status for a single node, so we
        will use the list API and find our node in there.

        :param bool invalidate_cache:
            (optional) if the cache should be invalidated or not

        :return:
            a status object or None if Node doesn't support status
        :rtype:
            `~ngenic.models.node_status.NodeStatus`
        """
        url = API_PATH["node_status"].format(tune_uuid=self._parent_tune_uuid)
        rsp_json = self._parse(await self._async_get(url, invalidate_cache))

        for status_obj in rsp_json:
            if status_obj["nodeUuid"] == self.uuid():
                return self._new_instance(NodeStatus, status_obj, node_uuid=self.uuid())
        return None



"""Room model for Ngenic Tune API."""

from typing import Any

import httpx

from ..const import API_PATH  # noqa: TID252
from .base import NgenicBase


class Room(NgenicBase):
    """Ngenic API room model."""

    def __init__(
        self, session: httpx.AsyncClient, json_data: dict[str, Any], tune_uuid: str
    ) -> None:
        """Initialize the room model."""
        self._parent_tune_uuid = tune_uuid

        super().__init__(session=session, json_data=json_data)

    def update(self) -> None:
        """Update this room with its current values."""

        url = API_PATH["rooms"].format(
            tune_uuid=self._parent_tune_uuid, room_uuid=self.uuid()
        )
        self._put(url, data=self.json())

    async def async_update(self) -> None:
        """Update this room with its current values (async)."""

        url = API_PATH["rooms"].format(
            tune_uuid=self._parent_tune_uuid, room_uuid=self.uuid()
        )
        await self._async_put(url, data=self.json())



"""Ngenic Node Status model."""

import logging
from typing import Any

import httpx

from .base import NgenicBase

_LOGGER = logging.getLogger(__name__)


class NodeStatus(NgenicBase):
    """Ngenic API node status model."""

    def __init__(
        self, session: httpx.AsyncClient, json_data: dict[str, Any], node_uuid: str
    ) -> None:
        """Initialize the node status model."""
        self._parent_node_uuid = node_uuid

        super().__init__(session=session, json_data=json_data)

    def battery_percentage(self) -> int:
        """Get the battery percentage."""
        if self["maxBattery"] == 0:
            # not using batteries
            _LOGGER.debug("Node %s is not using batteries", self._parent_node_uuid)
            return 100

        return int((self["battery"] / self["maxBattery"]) * 100)

    def radio_signal_percentage(self) -> int:
        """Get the radio signal percentage."""
        if self["maxRadioStatus"] == 0:
            # shouldn't happen as of now (always maxRadioStatus is always 4)
            return 100

        return int((self["radioStatus"] / self["maxRadioStatus"]) * 100)



"""A module initializing the Ngenic models package."""

from .base import NgenicBase  # noqa: F401
from .room import Room  # noqa: F401
from .setpoint_schedule import SetpointSchedule  # noqa: F401
from .tune import Tune  # noqa: F401
